// logsortts.cpp: Hauptprojektdatei.

// Eri's Log File Sorting Utility Ver. 0.0

#include "stdafx.h"

using namespace System;
using namespace System::Reflection;
using namespace System::IO;
using namespace System::Collections::Generic;

String ^AppVersion()
{
  String ^strVersion = Assembly::GetExecutingAssembly()->GetName()->Version->ToString();
#ifdef _DEBUG
  strVersion += " (debug)";
#endif
  return strVersion;
}

public ref class LogRecord : public IComparable<LogRecord ^>
{
public:
  static LogRecord ^FromLineArray(array<String ^> ^astrLines, int %nLineIndex)  // Updates the index passed as parameter
  {
    while (nLineIndex < astrLines->Length && String::IsNullOrWhiteSpace(astrLines[nLineIndex])) ++nLineIndex;
    if (nLineIndex >= astrLines->Length) return nullptr;
    String ^strCurrentLine = astrLines[nLineIndex];
    LogRecord ^lr = gcnew LogRecord;
    List<String ^> ^lstLines = gcnew List<String ^>;
    lstLines->Add(strCurrentLine);
    int nCharIndex = 0;
    int nCharEndIndex;
    if (strCurrentLine[nCharIndex] == '[') {  // Process ID
      ++nCharIndex;
      nCharEndIndex = strCurrentLine->IndexOfAny(gcnew array<Char>{']', ':'}, nCharIndex);
      lr->m_nProcessId = int::Parse(strCurrentLine->Substring(nCharIndex, nCharEndIndex - nCharIndex));
      if (strCurrentLine[nCharEndIndex] == ':') {  // Thread ID
        nCharIndex = nCharEndIndex + 1;
        nCharEndIndex = strCurrentLine->IndexOf(']', nCharIndex);
        lr->m_nThreadId = int::Parse(strCurrentLine->Substring(nCharIndex, nCharEndIndex - nCharIndex));
      }
      nCharIndex = nCharEndIndex + 2;
    }
    if (Char::IsDigit(strCurrentLine[nCharIndex])) {  // Timestamped line
      nCharEndIndex = strCurrentLine->IndexOf(": ", nCharIndex);
      if (nCharEndIndex == -1) nCharEndIndex = strCurrentLine->Length - 1;
      TryParseTimestamp(strCurrentLine->Substring(nCharIndex, nCharEndIndex - nCharIndex), lr->m_dtTimestamp);
      ++nLineIndex;
    } else {  // Default to header line block or termination line
      TryParseTimestamp(strCurrentLine->Substring(strCurrentLine->LastIndexOf(' ', strCurrentLine->LastIndexOf(' ') - 1) + 1),
        lr->m_dtTimestamp);
      if (strCurrentLine->Substring(nCharIndex)->StartsWith("Logging stopped")) {  // Optionally adjust timestamp for termination line
        if (!(lr->m_dtTimestamp.Ticks % TimeSpan::TicksPerSecond))  // Assume old-style termination timestamp if there's absolutely no seconds fraction
          lr->m_dtTimestamp = lr->m_dtTimestamp.AddSeconds(1).AddTicks(-1);
        lr->m_bAppendBlankLine = true;
        ++nLineIndex;
      } else  // Append the remaining header lines to this record
        while (astrLines[++nLineIndex][0] != '[' && !Char::IsDigit(astrLines[nLineIndex][0]))
          lstLines->Add(astrLines[nLineIndex]);
    }
    lr->m_astrLines = lstLines->ToArray();
    return lr;
  }

  virtual String ^ToString() override
  {
    String ^str = String::Join(Environment::NewLine, m_astrLines);
    if (m_bAppendBlankLine) str += Environment::NewLine;
    return str;
  }

  virtual int CompareTo(LogRecord ^other)
  {
    return m_dtTimestamp.CompareTo(other->m_dtTimestamp);
  }

private:
  static bool TryParseTimestamp(String ^s, DateTime %dtOut)
  {
    DateTime dt;
    TimeSpan tsTimeOfDay;
    if (TimeSpan::TryParse(s, tsTimeOfDay))  // First try to parse as TimeSpan (will fail if the timestamp contains a date)
      dt = s_dtCurrentDate + tsTimeOfDay;
    else  // Attempt to parse date plus time as one entity (will fail if the TOD contains fractional seconds)
      if (DateTime::TryParse(s, dt))
        s_dtCurrentDate = dt.Date;
    else {  // Finally attempt to split into date and TOD and parse separately
      array<String ^> ^as = s->Split(' ');
      if (as->Length >= 2 && DateTime::TryParse(as[0], dt) && TimeSpan::TryParse(as[1], tsTimeOfDay)) {
        s_dtCurrentDate = dt;
        dt += tsTimeOfDay;
      } else dt = DateTime::MinValue;  // To indicate failure
    }
    if (dt != DateTime::MinValue) {
      dtOut = dt;
      return true;
    } else return false;
  }

private:
  LogRecord() : m_nProcessId(-1), m_nThreadId(-1), m_dtTimestamp(DateTime::MinValue), m_bAppendBlankLine(false)
  {}

private:
  static DateTime s_dtCurrentDate = DateTime::MinValue;

  int m_nProcessId;
  int m_nThreadId;
  DateTime m_dtTimestamp;
  array<String ^> ^m_astrLines;
  bool m_bAppendBlankLine;
};

int main(array<System::String ^> ^args)
{
  Console::WriteLine("Eri's Log File Sorting Utility Ver. {0}", AppVersion());
  Console::WriteLine();
  if (args->Length < 1 || args->Length > 2) Console::WriteLine("Usage: logsortts <input filename> [<output filename>]");
  else {
    String ^strSortedUpToMarker = "[Log sorted by timestamp up to this point]";
    array<String ^> ^astrInputLines = File::ReadAllLines(args[0]);
    int nStartLineIndex = Array::LastIndexOf(astrInputLines, strSortedUpToMarker) + 1;
    List<LogRecord ^> ^lstLogRecords = gcnew List<LogRecord ^>;
    int i = nStartLineIndex;
    LogRecord ^lr;
    while (lr = LogRecord::FromLineArray(astrInputLines, i))
      lstLogRecords->Add(lr);
    lstLogRecords->Sort();
    Console::WriteLine("{0} log records sorted", lstLogRecords->Count);
    StreamWriter ^stwOutFile = File::CreateText(args[args->Length < 2 ? 0 : 1]);
    if (nStartLineIndex)
      for (i = 0; i < nStartLineIndex - 1; ++i) stwOutFile->WriteLine(astrInputLines[i]);
    for each (lr in lstLogRecords) stwOutFile->WriteLine(lr);
    stwOutFile->WriteLine(strSortedUpToMarker);
    stwOutFile->WriteLine();
    stwOutFile->Close();
  }
#ifdef _DEBUG
  Console::WriteLine();
  Console::WriteLine("Hit <Enter> to continue...");
  Console::ReadLine();
#endif
  return 0;
}
